
/*-
# X-BASED RUBIK'S CUBE(tm)
#
#  Rubik3d.c
#
###
#
#  Copyright (c) 1994 - 99	David Albert Bagley, bagleyd@tux.org
#
#                   All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "playable",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/* Methods file for Rubik3d */

#include <stdio.h>
#include <X11/IntrinsicP.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/CoreP.h>
#include "RubikP.h"
#include "Rubik3dP.h"

static void InitializeRubik3D(Widget request, Widget renew);
static void ExposeRubik3D(Widget renew, XEvent * event, Region region);
static void ResizeRubik3D(Rubik3DWidget w);
static Boolean SetValuesRubik3D(Widget current, Widget request, Widget renew);
static void MoveRubik3DTl(Rubik3DWidget w, XEvent * event, char **args, int nArgs);
static void MoveRubik3DTop(Rubik3DWidget w, XEvent * event, char **args, int nArgs);
static void MoveRubik3DTr(Rubik3DWidget w, XEvent * event, char **args, int nArgs);
static void MoveRubik3DLeft(Rubik3DWidget w, XEvent * event, char **args, int nArgs);
static void MoveRubik3DRight(Rubik3DWidget w, XEvent * event, char **args, int nArgs);
static void MoveRubik3DBl(Rubik3DWidget w, XEvent * event, char **args, int nArgs);
static void MoveRubik3DBottom(Rubik3DWidget w, XEvent * event, char **args, int nArgs);
static void MoveRubik3DBr(Rubik3DWidget w, XEvent * event, char **args, int nArgs);
static void ResizePolyhedrons(Rubik3DWidget w);
static void DrawFrame(Rubik3DWidget w, GC gc);
static void MapTo3D(Rubik3DWidget w, int face, int i, int j, int *x, int *y);

#ifdef DEBUG
static void MapFrom3D();

#endif
static void MapOrientFrom3D(int face, int corner, int *side);
static void CubeOffset3D(Rubik3DWidget w, int face, int x, int y, int *dx, int *dy);

static char defaultTranslationsRubik3D[] =
"<KeyPress>q: Quit()\n\
   Ctrl<KeyPress>C: Quit()\n\
   <KeyPress>KP_Divide: MoveCcw()\n\
   <KeyPress>Home: MoveTl()\n\
   <KeyPress>KP_7: MoveTl()\n\
   <KeyPress>R7: MoveTl()\n\
   <KeyPress>Up: MoveTop()\n\
   <KeyPress>KP_8: MoveTop()\n\
   <KeyPress>R8: MoveTop()\n\
   <KeyPress>Prior: MoveTr()\n\
   <KeyPress>KP_9: MoveTr()\n\
   <KeyPress>R9: MoveTr()\n\
   <KeyPress>Left: MoveLeft()\n\
   <KeyPress>KP_4: MoveLeft()\n\
   <KeyPress>R10: MoveLeft()\n\
   <KeyPress>Begin: MoveCw()\n\
   <KeyPress>KP_5: MoveCw()\n\
   <KeyPress>R11: MoveCw()\n\
   <KeyPress>Right: MoveRight()\n\
   <KeyPress>KP_6: MoveRight()\n\
   <KeyPress>R12: MoveRight()\n\
   <KeyPress>End: MoveBl()\n\
   <KeyPress>KP_1: MoveBl()\n\
   <KeyPress>R13: MoveBl()\n\
   <KeyPress>Down: MoveBottom()\n\
   <KeyPress>KP_2: MoveBottom()\n\
   <KeyPress>R14: MoveBottom()\n\
   <KeyPress>Next: MoveBr()\n\
   <KeyPress>KP_3: MoveBr()\n\
   <KeyPress>R15: MoveBr()\n\
   <Btn1Down>: Select()\n\
   <Btn1Up>: Release()\n\
   <KeyPress>p: Practice()\n\
   <Btn2Down>(2+): Practice()\n\
   <Btn2Down>: PracticeMaybe()\n\
   <KeyPress>r: Randomize()\n\
   <Btn3Down>(2+): Randomize()\n\
   <Btn3Down>: RandomizeMaybe()\n\
   <KeyPress>g: Get()\n\
   <KeyPress>w: Write()\n\
   <KeyPress>u: Undo()\n\
   <KeyPress>s: Solve()\n\
   <KeyPress>d: Decrement()\n\
   <KeyPress>i: Increment()\n\
   <KeyPress>x: IncrementX()\n\
   <KeyPress>y: IncrementY()\n\
   <KeyPress>z: IncrementZ()\n\
   <KeyPress>o: Orientize()";

static XtActionsRec actionsListRubik3D[] =
{
	{"Quit", (XtActionProc) QuitRubik},
	{"MoveCcw", (XtActionProc) MoveRubikCcw},
	{"MoveTl", (XtActionProc) MoveRubik3DTl},
	{"MoveTop", (XtActionProc) MoveRubik3DTop},
	{"MoveTr", (XtActionProc) MoveRubik3DTr},
	{"MoveLeft", (XtActionProc) MoveRubik3DLeft},
	{"MoveCw", (XtActionProc) MoveRubikCw},
	{"MoveRight", (XtActionProc) MoveRubik3DRight},
	{"MoveBl", (XtActionProc) MoveRubik3DBl},
	{"MoveBottom", (XtActionProc) MoveRubik3DBottom},
	{"MoveBr", (XtActionProc) MoveRubik3DBr},
	{"Select", (XtActionProc) SelectRubik},
	{"Release", (XtActionProc) ReleaseRubik},
	{"Practice", (XtActionProc) PracticeRubik},
	{"PracticeMaybe", (XtActionProc) PracticeRubikMaybe},
	{"Randomize", (XtActionProc) RandomizeRubik},
	{"RandomizeMaybe", (XtActionProc) RandomizeRubikMaybe},
	{"Get", (XtActionProc) GetRubik},
	{"Write", (XtActionProc) WriteRubik},
	{"Undo", (XtActionProc) UndoRubik},
	{"Solve", (XtActionProc) SolveRubik},
	{"Decrement", (XtActionProc) DecrementRubik},
	{"Increment", (XtActionProc) IncrementRubik},
	{"IncrementX", (XtActionProc) IncrementXRubik},
	{"IncrementY", (XtActionProc) IncrementYRubik},
	{"IncrementZ", (XtActionProc) IncrementZRubik},
	{"Orientize", (XtActionProc) OrientizeRubik}
};

static XtResource resourcesRubik3D[] =
{
	{XtNuserName, XtCUserName, XtRString, sizeof (String),
	 XtOffset(RubikWidget, rubik.username), XtRString, "nobody"},
	{XtNfaceColor0, XtCLabel, XtRString, sizeof (String),
	 XtOffset(RubikWidget, rubik.faceName[0]), XtRString, "Red"},
	{XtNfaceColor1, XtCLabel, XtRString, sizeof (String),
	 XtOffset(RubikWidget, rubik.faceName[1]), XtRString, "Yellow"},
	{XtNfaceColor2, XtCLabel, XtRString, sizeof (String),
	 XtOffset(RubikWidget, rubik.faceName[2]), XtRString, "White"},
	{XtNfaceColor3, XtCLabel, XtRString, sizeof (String),
	 XtOffset(RubikWidget, rubik.faceName[3]), XtRString, "Green"},
	{XtNfaceColor4, XtCLabel, XtRString, sizeof (String),
	 XtOffset(RubikWidget, rubik.faceName[4]), XtRString, "Orange"},
	{XtNfaceColor5, XtCLabel, XtRString, sizeof (String),
	 XtOffset(RubikWidget, rubik.faceName[5]), XtRString, "Blue"},
	{XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
    XtOffset(RubikWidget, rubik.foreground), XtRString, XtDefaultForeground},
	{XtNpieceBorder, XtCColor, XtRPixel, sizeof (Pixel),
   XtOffset(RubikWidget, rubik.borderColor), XtRString, XtDefaultForeground},
	{XtNwidth, XtCWidth, XtRDimension, sizeof (Dimension),
	 XtOffset(RubikWidget, core.width), XtRString, "250"},
	{XtNheight, XtCHeight, XtRDimension, sizeof (Dimension),
	 XtOffset(RubikWidget, core.height), XtRString, "400"},
	{XtNsizex, XtCSizeX, XtRInt, sizeof (int),
	 XtOffset(RubikWidget, rubik.sizex), XtRString, "3"},	/* DEFAULTCUBES */
	{XtNsizey, XtCSizeY, XtRInt, sizeof (int),
	 XtOffset(RubikWidget, rubik.sizey), XtRString, "3"},	/* DEFAULTCUBES */
	{XtNsizez, XtCSizeZ, XtRInt, sizeof (int),
	 XtOffset(RubikWidget, rubik.sizez), XtRString, "3"},	/* DEFAULTCUBES */
	{XtNorient, XtCOrient, XtRBoolean, sizeof (Boolean),
	 XtOffset(RubikWidget, rubik.orient), XtRString, "FALSE"},	/* DEFAULTORIENT */
	{XtNmono, XtCMono, XtRBoolean, sizeof (Boolean),
	 XtOffset(RubikWidget, rubik.mono), XtRString, "FALSE"},
	{XtNreverse, XtCReverse, XtRBoolean, sizeof (Boolean),
	 XtOffset(RubikWidget, rubik.reverse), XtRString, "FALSE"},
	{XtNface, XtCFace, XtRInt, sizeof (int),
	 XtOffset(RubikWidget, rubik.currentFace), XtRString, "-1"},
	{XtNpos, XtCPos, XtRInt, sizeof (int),
	 XtOffset(RubikWidget, rubik.currentPosition), XtRString, "-1"},
	{XtNdirection, XtCDirection, XtRInt, sizeof (int),
	 XtOffset(RubikWidget, rubik.currentDirection), XtRString, "-1"},
	{XtNpractice, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(RubikWidget, rubik.practice), XtRString, "FALSE"},
	{XtNstart, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(RubikWidget, rubik.started), XtRString, "FALSE"},
	{XtNselectCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
	 XtOffset(RubikWidget, rubik.select), XtRCallback, NULL}
};

Rubik3DClassRec rubik3dClassRec =
{
	{
		(WidgetClass) & rubikClassRec,	/* superclass */
		"Rubik3D",	/* class name */
		sizeof (Rubik3DRec),	/* widget size */
		NULL,		/* class initialize */
		NULL,		/* class part initialize */
		FALSE,		/* class inited */
		(XtInitProc) InitializeRubik3D,		/* initialize */
		NULL,		/* initialize hook */
		XtInheritRealize,	/* realize */
		actionsListRubik3D,	/* actions */
		XtNumber(actionsListRubik3D),	/* num actions */
		resourcesRubik3D,	/* resources */
		XtNumber(resourcesRubik3D),	/* num resources */
		NULLQUARK,	/* xrm class */
		TRUE,		/* compress motion */
		TRUE,		/* compress exposure */
		TRUE,		/* compress enterleave */
		TRUE,		/* visible interest */
		NULL,		/* destroy */
		(XtWidgetProc) ResizeRubik3D,	/* resize */
		(XtExposeProc) ExposeRubik3D,	/* expose */
		(XtSetValuesFunc) SetValuesRubik3D,	/* set values */
		NULL,		/* set values hook */
		XtInheritSetValuesAlmost,	/* set values almost */
		NULL,		/* get values hook */
		XtInheritAcceptFocus,	/* accept focus */
		XtVersion,	/* version */
		NULL,		/* callback private */
		defaultTranslationsRubik3D,	/* tm table */
		NULL,		/* query geometry */
		NULL,		/* display accelerator */
		NULL		/* extension */
	},
	{
		0		/* ignore */
	},
	{
		0		/* ignore */
	}
};

WidgetClass rubik3dWidgetClass = (WidgetClass) & rubik3dClassRec;

static XPoint faceLoc3D[MAXFACES][MAXORIENT];
static XPoint cubeLoc3D[MAXFACES][MAXORIENT + 1];
static XPoint letter3DList[MAXFACES];
static XPoint orient3DList[MAXFACES][MAXORIENT][2];
static RowNext rotateToRow[MAXFACES] =	/*CW to min face */
{
	{1, LEFT, TOP},
	{0, BOTTOM, RIGHT},
	{0, RIGHT, BOTTOM},
	{0, TOP, LEFT},
	{1, RIGHT, BOTTOM},
	{0, LEFT, TOP}
};

static void
InitializeRubik3D(Widget request, Widget renew)
{
	Rubik3DWidget w = (Rubik3DWidget) renew;

	w->rubik.dim = 3;
	ResizeRubik3D(w);
}

static void
ResizeRubik3D(Rubik3DWidget w)
{
	/* XPoint      tempSize; */

	w->rubik.delta = 4;
	w->rubik.vertical = (w->core.height >= w->core.width);
	/* Unlike the 2d implementation... lets have variable size cubies and
	   static sized cube */
	if (w->rubik.vertical) {
		w->rubik3d.viewSize.y = w->core.height / MAXVIEWS;
		w->rubik3d.viewSize.x = w->core.width;
		if (w->rubik3d.viewSize.x >= DIVIDE(w->rubik3d.viewSize.y)) {
			w->rubik3d.viewSize.y = MAX(w->rubik3d.viewSize.y, 0);
			w->rubik3d.viewSize.x = DIVIDE(w->rubik3d.viewSize.y);
		} else {
			w->rubik3d.viewSize.x = MAX(w->rubik3d.viewSize.x, 0);
			w->rubik3d.viewSize.y = MULTIPLY(w->rubik3d.viewSize.x);
		}
		w->rubik3d.faceSize.x = w->rubik3d.viewSize.x / 2 - 2;
		w->rubik3d.faceSize.y = w->rubik3d.viewSize.y / 2 - 2;
		w->rubik3d.faceDiagonal = w->rubik3d.faceSize.x / 2;
		w->rubik.puzzleSize.x = w->rubik3d.viewSize.x;
		w->rubik.puzzleSize.y = MAXVIEWS * w->rubik3d.viewSize.y;
	} else {
		w->rubik3d.viewSize.x = w->core.width / MAXVIEWS;
		w->rubik3d.viewSize.y = w->core.height;
		if (w->rubik3d.viewSize.y >= DIVIDE(w->rubik3d.viewSize.x)) {
			w->rubik3d.viewSize.x = MAX(w->rubik3d.viewSize.x, 0);
			w->rubik3d.viewSize.y = DIVIDE(w->rubik3d.viewSize.x);
		} else {
			w->rubik3d.viewSize.y = MAX(w->rubik3d.viewSize.y, 0);
			w->rubik3d.viewSize.x = MULTIPLY(w->rubik3d.viewSize.y);
		}
		w->rubik3d.faceSize.x = w->rubik3d.viewSize.x / 2 - 2;
		w->rubik3d.faceSize.y = w->rubik3d.viewSize.y / 2 - 2;
		w->rubik3d.faceDiagonal = w->rubik3d.faceSize.y / 2;
		w->rubik.puzzleSize.y = w->rubik3d.viewSize.y;
		w->rubik.puzzleSize.x = MAXVIEWS * w->rubik3d.viewSize.x;
	}
	w->rubik3d.cubeSizex.x = MAX(w->rubik3d.faceSize.x /
				     w->rubik.sizex - w->rubik.delta, 0);
	w->rubik3d.cubeSizey.x = MAX(w->rubik3d.faceSize.x /
				     w->rubik.sizey - w->rubik.delta, 0);
	w->rubik3d.cubeSizez.x = MAX(w->rubik3d.faceSize.x /
				     w->rubik.sizez - w->rubik.delta, 0);
	w->rubik3d.cubeSizex.y = MAX(w->rubik3d.faceSize.y /
				     w->rubik.sizex - w->rubik.delta, 0);
	w->rubik3d.cubeSizey.y = MAX(w->rubik3d.faceSize.y /
				     w->rubik.sizey - w->rubik.delta, 0);
	w->rubik3d.cubeSizez.y = MAX(w->rubik3d.faceSize.y /
				     w->rubik.sizez - w->rubik.delta, 0);
	w->rubik.puzzleOffset.x = ((int) w->core.width - w->rubik.puzzleSize.x) / 2;
	w->rubik.puzzleOffset.y = ((int) w->core.height - w->rubik.puzzleSize.y) /
		2;
	ResizePolyhedrons(w);
}

static void
ExposeRubik3D(Widget renew, XEvent * event, Region region)
{
	Rubik3DWidget w = (Rubik3DWidget) renew;

	if (w->core.visible) {
		if (w->rubik.reverse)
			XFillRectangle(XtDisplay(w), XtWindow(w),
				       w->rubik.inverseGC, 0, 0, w->core.width, w->core.height);
		DrawFrame(w, w->rubik.puzzleGC);
		DrawAllPolyhedrons((RubikWidget) w);
	}
}

static      Boolean
SetValuesRubik3D(Widget current, Widget request, Widget renew)
{
	Rubik3DWidget c = (Rubik3DWidget) current, w = (Rubik3DWidget) renew;
	Boolean     redraw = False;

	if (w->rubik.sizex != c->rubik.sizex ||
	    w->rubik.sizey != c->rubik.sizey ||
	    w->rubik.sizez != c->rubik.sizez) {
		ResetPolyhedrons((RubikWidget) w);
		ResizeRubik3D(w);
		redraw = True;
	}
	if (w->rubik3d.cubeSizex.x != c->rubik3d.cubeSizex.x ||
	    w->rubik3d.cubeSizey.x != c->rubik3d.cubeSizey.x ||
	    w->rubik3d.cubeSizez.x != c->rubik3d.cubeSizez.x ||
	    w->rubik3d.cubeSizex.y != c->rubik3d.cubeSizex.y ||
	    w->rubik3d.cubeSizey.y != c->rubik3d.cubeSizey.y ||
	    w->rubik3d.cubeSizez.y != c->rubik3d.cubeSizez.y) {
		ResizeRubik3D(w);
		redraw = True;
	}
	return (redraw);
}

static void
MoveRubik3DTl(Rubik3DWidget w, XEvent * event, char **args, int nArgs)
{
	MoveRubikInput((RubikWidget) w, event->xbutton.x, event->xbutton.y, TL,
		       (int) (event->xkey.state & ControlMask));
}

static void
MoveRubik3DTop(Rubik3DWidget w, XEvent * event, char **args, int nArgs)
{
	MoveRubikInput((RubikWidget) w, event->xbutton.x, event->xbutton.y, TOP,
		       (int) (event->xkey.state & ControlMask));
}

static void
MoveRubik3DTr(Rubik3DWidget w, XEvent * event, char **args, int nArgs)
{
	MoveRubikInput((RubikWidget) w, event->xbutton.x, event->xbutton.y, TR,
		       (int) (event->xkey.state & ControlMask));
}

static void
MoveRubik3DLeft(Rubik3DWidget w, XEvent * event, char **args, int nArgs)
{
	MoveRubikInput((RubikWidget) w, event->xbutton.x, event->xbutton.y, LEFT,
		       (int) (event->xkey.state & ControlMask));
}

static void
MoveRubik3DRight(Rubik3DWidget w, XEvent * event, char **args, int nArgs)
{
	MoveRubikInput((RubikWidget) w, event->xbutton.x, event->xbutton.y, RIGHT,
		       (int) (event->xkey.state & ControlMask));
}

static void
MoveRubik3DBl(Rubik3DWidget w, XEvent * event, char **args, int nArgs)
{
	MoveRubikInput((RubikWidget) w, event->xbutton.x, event->xbutton.y, BL,
		       (int) (event->xkey.state & ControlMask));
}

static void
MoveRubik3DBottom(Rubik3DWidget w, XEvent * event, char **args, int nArgs)
{
	MoveRubikInput((RubikWidget) w, event->xbutton.x, event->xbutton.y, BOTTOM,
		       (int) (event->xkey.state & ControlMask));
}

static void
MoveRubik3DBr(Rubik3DWidget w, XEvent * event, char **args, int nArgs)
{
	MoveRubikInput((RubikWidget) w, event->xbutton.x, event->xbutton.y, BR,
		       (int) (event->xkey.state & ControlMask));
}

static void
ResizePolyhedrons(Rubik3DWidget w)
{
	int         face, orient, side, corner, sizex, sizey;
	XPoint      subcubeLoc3D[MAXFACES][MAXORIENT];
	XPoint      diamondLoc3D[MAXFACES][MAXORIENT];
	XPoint      subdiamondLoc3D[MAXFACES][MAXORIENT];

	w->rubik.letterOffset.x = -2;
	w->rubik.letterOffset.y = 4;
	w->rubik3d.viewMiddle.x = w->rubik3d.faceSize.x +
		w->rubik.puzzleOffset.x;
	w->rubik3d.viewMiddle.y = w->rubik3d.faceSize.y +
		w->rubik.puzzleOffset.y;
	for (face = 0; face < MAXFACES; face++) {
		faceLoc3D[face][0].x = w->rubik3d.viewMiddle.x;
		faceLoc3D[face][0].y = w->rubik3d.viewMiddle.y;
		for (orient = 1; orient < MAXORIENT; orient++) {
			faceLoc3D[face][orient].x = w->rubik3d.faceSize.x;
			faceLoc3D[face][orient].y = w->rubik3d.faceSize.y;
		}
	}
	if (w->rubik.vertical) {
		faceLoc3D[0][1].x /= -2;
		faceLoc3D[0][1].y /= -1;
		faceLoc3D[0][2].y = 0;
		faceLoc3D[0][3].x /= 2;

		faceLoc3D[1][1].x /= -2;
		faceLoc3D[1][2].x /= -2;
		faceLoc3D[1][2].y /= -1;
		faceLoc3D[1][3].x /= 2;
		faceLoc3D[1][3].y /= -1;

		faceLoc3D[2][1].y = 0;
		faceLoc3D[2][2].x /= -2;
		faceLoc3D[2][3].x /= -1;
		faceLoc3D[2][3].y = 0;

		for (face = MAXFACES / 2; face < MAXFACES; face++)
			faceLoc3D[face][0].y += w->rubik3d.viewSize.y + 3;

		faceLoc3D[3][1].x /= 2;
		faceLoc3D[3][1].y /= -1;
		faceLoc3D[3][2].x /= 2;
		faceLoc3D[3][3].x /= -2;

		faceLoc3D[4][1].x /= -1;
		faceLoc3D[4][1].y = 0;
		faceLoc3D[4][2].x /= 2;
		faceLoc3D[4][2].y /= -1;
		faceLoc3D[4][3].y = 0;

		faceLoc3D[5][1].x /= 2;
		faceLoc3D[5][2].x /= -1;
		faceLoc3D[5][2].y = 0;
		faceLoc3D[5][3].x /= -2;
		faceLoc3D[5][3].y /= -1;
	} else {
		faceLoc3D[0][1].x /= -1;
		faceLoc3D[0][1].y /= -2;
		faceLoc3D[0][2].y /= -2;
		faceLoc3D[0][3].y /= 2;

		faceLoc3D[1][1].x = 0;
		faceLoc3D[1][2].x /= -1;
		faceLoc3D[1][2].y /= -2;
		faceLoc3D[1][3].x = 0;
		faceLoc3D[1][3].y /= -1;

		faceLoc3D[2][1].y /= -2;
		faceLoc3D[2][2].x = 0;
		faceLoc3D[2][3].y /= 2;
		faceLoc3D[2][3].x /= -1;

		for (face = MAXFACES / 2; face < MAXFACES; face++)
			faceLoc3D[face][0].x += w->rubik3d.viewSize.x + 3;

		faceLoc3D[3][1].x /= -1;
		faceLoc3D[3][1].y /= 2;
		faceLoc3D[3][2].x = 0;
		faceLoc3D[3][2].y /= -1;
		faceLoc3D[3][3].y /= -2;

		faceLoc3D[4][1].y /= 2;
		faceLoc3D[4][2].x /= -1;
		faceLoc3D[4][2].y /= 2;
		faceLoc3D[4][3].x /= -1;
		faceLoc3D[4][3].y /= -2;

		faceLoc3D[5][1].x = 0;
		faceLoc3D[5][1].y /= -1;
		faceLoc3D[5][2].y /= 2;
		faceLoc3D[5][3].x = 0;
	}
	for (face = 0; face < MAXFACES; face++) {
		cubeLoc3D[face][0].x = faceLoc3D[face][0].x;
		cubeLoc3D[face][0].y = faceLoc3D[face][0].y;
		subcubeLoc3D[face][0].x = faceLoc3D[face][0].x;
		subcubeLoc3D[face][0].y = faceLoc3D[face][0].y;
		if (face == 0 || face == 4) {
			sizex = w->rubik.sizex;
			sizey = w->rubik.sizez;
		} else if (face == 1 || face == 3) {
			sizex = w->rubik.sizez;
			sizey = w->rubik.sizey;
		} else {	/* if (face == 2 || face == 5) */
			sizex = w->rubik.sizex;
			sizey = w->rubik.sizey;
		}
		for (orient = 1; orient < MAXORIENT; orient++) {
			if (((orient & 1) && (face == 2 || face == 3 || face == 4)) ||
			    (!(orient & 1) && !(face == 2 || face == 3 || face == 4)))
				cubeLoc3D[face][orient].x =
					(faceLoc3D[face][orient].x - (3 + sizex) *
				 w->rubik.delta * faceLoc3D[face][orient].x /
					 w->rubik3d.faceSize.x) / sizex;
			else
				cubeLoc3D[face][orient].x =
					(faceLoc3D[face][orient].x - (3 + sizey) *
				 w->rubik.delta * faceLoc3D[face][orient].x /
					 w->rubik3d.faceSize.x) / sizey;
			if (((orient & 1) && !(face == 2 || face == 3 || face == 4)) ||
			    (!(orient & 1) && (face == 2 || face == 3 || face == 4)))
				cubeLoc3D[face][orient].y =
					(faceLoc3D[face][orient].y - (3 + sizey) *
				 w->rubik.delta * faceLoc3D[face][orient].y /
					 w->rubik3d.faceSize.y) / sizey;
			else
				cubeLoc3D[face][orient].y =
					(faceLoc3D[face][orient].y - (3 + sizex) *
				 w->rubik.delta * faceLoc3D[face][orient].y /
					 w->rubik3d.faceSize.y) / sizex;
			if (((orient & 1) && (face == 2 || face == 3 || face == 4)) ||
			    (!(orient & 1) && !(face == 2 || face == 3 || face == 4)))
				subcubeLoc3D[face][orient].x =
					(faceLoc3D[face][orient].x - (5 + sizex) *
				 w->rubik.delta * faceLoc3D[face][orient].x /
					 w->rubik3d.faceSize.x) / (2 * sizex);
			else
				subcubeLoc3D[face][orient].x =
					(faceLoc3D[face][orient].x - (5 + sizey) *
				 w->rubik.delta * faceLoc3D[face][orient].x /
					 w->rubik3d.faceSize.x) / (2 * sizey);
			if (((orient & 1) && !(face == 2 || face == 3 || face == 4)) ||
			    (!(orient & 1) && (face == 2 || face == 3 || face == 4)))
				subcubeLoc3D[face][orient].y =
					(faceLoc3D[face][orient].y - (5 + sizey) *
				 w->rubik.delta * faceLoc3D[face][orient].y /
					 w->rubik3d.faceSize.y) / (2 * sizey);
			else
				subcubeLoc3D[face][orient].y =
					(faceLoc3D[face][orient].y - (5 + sizex) *
				 w->rubik.delta * faceLoc3D[face][orient].y /
					 w->rubik3d.faceSize.y) / (2 * sizex);
		}
		cubeLoc3D[face][MAXORIENT].x = -cubeLoc3D[face][1].x -
			cubeLoc3D[face][2].x - cubeLoc3D[face][3].x;
		cubeLoc3D[face][MAXORIENT].y = -cubeLoc3D[face][1].y -
			cubeLoc3D[face][2].y - cubeLoc3D[face][3].y;
	}
	w->rubik3d.cubeSizex.x = MAX(w->rubik3d.faceSize.x /
				     w->rubik.sizex - w->rubik.delta, 0);
	w->rubik3d.cubeSizey.x = MAX(w->rubik3d.faceSize.x /
				     w->rubik.sizey - w->rubik.delta, 0);
	w->rubik3d.cubeSizez.x = MAX(w->rubik3d.faceSize.x /
				     w->rubik.sizez - w->rubik.delta, 0);
	w->rubik3d.cubeSizex.y = MAX(w->rubik3d.faceSize.y /
				     w->rubik.sizex - w->rubik.delta, 0);
	w->rubik3d.cubeSizey.y = MAX(w->rubik3d.faceSize.y /
				     w->rubik.sizey - w->rubik.delta, 0);
	w->rubik3d.cubeSizez.y = MAX(w->rubik3d.faceSize.y /
				     w->rubik.sizez - w->rubik.delta, 0);
	w->rubik3d.cubeDiagonalx = (w->rubik3d.faceDiagonal - w->rubik.delta) /
		w->rubik.sizex - w->rubik.delta;
	w->rubik3d.cubeDiagonaly = (w->rubik3d.faceDiagonal - w->rubik.delta) /
		w->rubik.sizey - w->rubik.delta;
	w->rubik3d.cubeDiagonalz = (w->rubik3d.faceDiagonal - w->rubik.delta) /
		w->rubik.sizez - w->rubik.delta;
	if (w->rubik.vertical) {
		letter3DList[0].x = w->rubik3d.cubeSizex.x / 4;
		letter3DList[0].y = -w->rubik3d.cubeSizez.y / 2;
		letter3DList[1].x = -w->rubik3d.cubeDiagonalz;
		letter3DList[1].y = 0;
		letter3DList[2].x = w->rubik3d.cubeSizex.x / 4;
		letter3DList[2].y = w->rubik3d.cubeSizey.y / 2;
		letter3DList[3].x = w->rubik3d.cubeDiagonalz;
		letter3DList[3].y = 0;
		letter3DList[4].x = -w->rubik3d.cubeSizex.x / 4;
		letter3DList[4].y = -w->rubik3d.cubeSizez.y / 2;
		letter3DList[5].x = -w->rubik3d.cubeSizex.x / 4;
		letter3DList[5].y = w->rubik3d.cubeSizey.y / 2;
	} else {
		letter3DList[0].x = 0;
		letter3DList[0].y = -w->rubik3d.cubeDiagonalz;
		letter3DList[1].x = -w->rubik3d.cubeSizez.x / 2;
		letter3DList[1].y = w->rubik3d.cubeSizey.y / 4;
		letter3DList[2].x = w->rubik3d.cubeSizex.x / 2;
		letter3DList[2].y = w->rubik3d.cubeSizey.y / 4;
		letter3DList[3].x = -w->rubik3d.cubeSizez.x / 2;
		letter3DList[3].y = -w->rubik3d.cubeSizey.y / 4;
		letter3DList[4].x = 0;
		letter3DList[4].y = w->rubik3d.cubeDiagonalz;
		letter3DList[5].x = w->rubik3d.cubeSizex.x / 2;
		letter3DList[5].y = -w->rubik3d.cubeSizey.y / 4;
	}
	/* The following figures out where to put the orient lines */
	for (face = 0; face < MAXFACES; face++) {
		for (orient = 0; orient < MAXORIENT - 1; orient++) {
			diamondLoc3D[face][orient].x = (cubeLoc3D[face][orient].x +
					  cubeLoc3D[face][orient + 1].x) / 2;
			diamondLoc3D[face][orient].y = (cubeLoc3D[face][orient].y +
					  cubeLoc3D[face][orient + 1].y) / 2;
			subdiamondLoc3D[face][orient].x = (subcubeLoc3D[face][orient].x +
				       subcubeLoc3D[face][orient + 1].x) / 2;
			subdiamondLoc3D[face][orient].y = (subcubeLoc3D[face][orient].y +
				       subcubeLoc3D[face][orient + 1].y) / 2;
		}
		/* Its a parallelagram so take advantage of that */
		diamondLoc3D[face][orient].x = (cubeLoc3D[face][MAXORIENT - 1].x -
				       cubeLoc3D[face][MAXORIENT / 2].x) / 2;
		diamondLoc3D[face][orient].y = (cubeLoc3D[face][MAXORIENT - 1].y -
				       cubeLoc3D[face][MAXORIENT / 2].y) / 2;
		subdiamondLoc3D[face][orient].x = (subcubeLoc3D[face][MAXORIENT - 1].x -
				    subcubeLoc3D[face][MAXORIENT / 2].x) / 2;
		subdiamondLoc3D[face][orient].y = (subcubeLoc3D[face][MAXORIENT - 1].y -
				    subcubeLoc3D[face][MAXORIENT / 2].y) / 2;

		MapOrientFrom3D(face, 1, &corner);
		orient3DList[face][corner][0].x = cubeLoc3D[face][1].x / 2;
		orient3DList[face][corner][0].y = cubeLoc3D[face][1].y / 2;
		orient3DList[face][corner][1].x = orient3DList[face][corner][0].x +
			(cubeLoc3D[face][2].x - subcubeLoc3D[face][2].x) / 2;
		orient3DList[face][corner][1].y = orient3DList[face][corner][0].y +
			(cubeLoc3D[face][2].y - subcubeLoc3D[face][2].y) / 2;
		for (orient = 1; orient < MAXORIENT; orient++) {
			side = corner;
			MapOrientFrom3D(face, (orient + 1) % MAXORIENT, &corner);
			orient3DList[face][corner][0].x =
				orient3DList[face][side][0].x + diamondLoc3D[face][orient].x;
			orient3DList[face][corner][0].y =
				orient3DList[face][side][0].y + diamondLoc3D[face][orient].y;
			orient3DList[face][corner][1].x =
				orient3DList[face][side][1].x + subdiamondLoc3D[face][orient].x;
			orient3DList[face][corner][1].y =
				orient3DList[face][side][1].y + subdiamondLoc3D[face][orient].y;
		}
	}
}

Boolean
SelectPolyhedrons3D(Rubik3DWidget w, int x, int y, int *face, int *position)
{
	int         u, v, front, tl, ur, ul, i, j, sizeOfRow, sizeOfColumn;

	if (w->rubik.vertical) {
		x -= w->rubik3d.viewMiddle.x;
		front = (y < w->rubik3d.viewSize.y + w->rubik.puzzleOffset.y);
		if (!front)
			y -= (w->rubik3d.viewSize.y);
		tl = (y < w->rubik3d.viewMiddle.y);
		y -= w->rubik3d.viewMiddle.y;
		u = -w->rubik3d.faceSize.y * x + w->rubik3d.faceDiagonal * y;
		v = w->rubik3d.faceSize.y * x + w->rubik3d.faceDiagonal * y;
		ur = (u < 0);
		ul = (v < 0);
		if (front) {
			if (tl)
				*face = (ur) ? 0 : 1;
			else
				*face = (ul) ? 1 : 2;
		} else {
			if (tl)
				*face = (ul) ? 4 : 3;
			else
				*face = (ur) ? 3 : 5;
		}
	} else {
		y -= w->rubik3d.viewMiddle.y;
		front = (x < w->rubik3d.viewSize.x + w->rubik.puzzleOffset.x);
		if (!front)
			x -= (w->rubik3d.viewSize.x);
		tl = (x < w->rubik3d.viewMiddle.x);
		x -= w->rubik3d.viewMiddle.x;
		u = -w->rubik3d.faceSize.x * y + w->rubik3d.faceDiagonal * x;
		v = w->rubik3d.faceSize.x * y + w->rubik3d.faceDiagonal * x;
		ur = (u < 0);
		ul = (v < 0);
		if (front) {
			if (tl)
				*face = (ur) ? 1 : 0;
			else
				*face = (ul) ? 0 : 2;
		} else {
			if (tl)
				*face = (ul) ? 3 : 4;
			else
				*face = (ur) ? 4 : 5;
		}
	}
	faceSizes((RubikWidget) w, *face, &sizeOfRow, &sizeOfColumn);
	if (w->rubik.vertical)
		switch (*face) {
			case 0:
				i = (x - 2 - (y * w->rubik3d.faceDiagonal / w->rubik3d.faceSize.y)) /
					(w->rubik3d.cubeSizex.x + w->rubik.delta);
				j = (y + 2) / (w->rubik3d.cubeSizez.y + w->rubik.delta) +
					sizeOfColumn - 1;
				break;
			case 1:
				i = (x + 4 + (y * w->rubik3d.faceDiagonal / w->rubik3d.faceSize.y)) /
					(2 * (w->rubik3d.cubeDiagonalz + w->rubik.delta)) + sizeOfRow - 1;
				j = (-x - 6 + (y * w->rubik3d.faceDiagonal / w->rubik3d.faceSize.y)) /
					(2 * (w->rubik3d.cubeDiagonaly + w->rubik.delta));
				break;
			case 2:
				i = (x - 4 + (y * w->rubik3d.faceDiagonal / w->rubik3d.faceSize.y)) /
					(w->rubik3d.cubeSizex.x + w->rubik.delta);
				j = (y - 4) / (w->rubik3d.cubeSizey.y + w->rubik.delta);
				break;
			case 3:
				i = (-x + 5 + (y * w->rubik3d.faceDiagonal / w->rubik3d.faceSize.y)) /
					(2 * (w->rubik3d.cubeDiagonalz + w->rubik.delta)) + sizeOfRow - 1;
				j = (-x + 7 - (y * w->rubik3d.faceDiagonal / w->rubik3d.faceSize.y)) /
					(2 * (w->rubik3d.cubeDiagonaly + w->rubik.delta)) + sizeOfColumn - 1;
				break;
			case 4:
				i = (x + (y * w->rubik3d.faceDiagonal / w->rubik3d.faceSize.y)) /
					(w->rubik3d.cubeSizex.x + w->rubik.delta) + sizeOfRow - 1;
				j = (y - 2) / (w->rubik3d.cubeSizez.y + w->rubik.delta) +
					sizeOfColumn - 1;
				break;
			case 5:
				i = (x + 2 - (y * w->rubik3d.faceDiagonal / w->rubik3d.faceSize.y)) /
					(w->rubik3d.cubeSizex.x + w->rubik.delta) + sizeOfRow - 1;
				j = (y - 6) / (w->rubik3d.cubeSizey.y + w->rubik.delta);
				break;
			default:
				return False;
	} else
		switch (*face) {
			case 0:
				i = (-y - 3 + (x * w->rubik3d.faceDiagonal / w->rubik3d.faceSize.x)) /
					(2 * (w->rubik3d.cubeDiagonalx + w->rubik.delta));
				j = (y + 1 + (x * w->rubik3d.faceDiagonal / w->rubik3d.faceSize.x)) /
					(2 * (w->rubik3d.cubeDiagonalz + w->rubik.delta)) + sizeOfColumn - 1;
				break;
			case 1:
				i = (x + 2) / (w->rubik3d.cubeSizez.x + w->rubik.delta) +
					sizeOfRow - 1;
				j = (y - 3 - (x * w->rubik3d.faceDiagonal / w->rubik3d.faceSize.x)) /
					(w->rubik3d.cubeSizey.y + w->rubik.delta);
				break;
			case 2:
				i = (x - 4) / (w->rubik3d.cubeSizex.x + w->rubik.delta);
				j = (y - 6 + (x * w->rubik3d.faceDiagonal / w->rubik3d.faceSize.x)) /
					(w->rubik3d.cubeSizey.y + w->rubik.delta);
				break;
			case 3:
				i = x / (w->rubik3d.cubeSizez.x + w->rubik.delta) +
					sizeOfRow - 1;
				j = (y + (x * w->rubik3d.faceDiagonal / w->rubik3d.faceSize.x)) /
					(w->rubik3d.cubeSizey.y + w->rubik.delta) + sizeOfColumn - 1;
				break;
			case 4:
				i = (-y + 9 - (x * w->rubik3d.faceDiagonal / w->rubik3d.faceSize.x)) /
					(2 * (w->rubik3d.cubeDiagonalx + w->rubik.delta)) + sizeOfRow - 1;
				j = (-y + 7 + (x * w->rubik3d.faceDiagonal / w->rubik3d.faceSize.x)) /
					(2 * (w->rubik3d.cubeDiagonalz + w->rubik.delta)) + sizeOfColumn - 1;
				break;
			case 5:
				i = (-x + 7) / (w->rubik3d.cubeSizex.x + w->rubik.delta) +
					sizeOfRow - 1;
				j = (-y - 4 + (x * w->rubik3d.faceDiagonal / w->rubik3d.faceSize.x)) /
					(w->rubik3d.cubeSizey.y + w->rubik.delta);
				break;
			default:
				return False;
		}
	if (i < 0 || j < 0 || i >= sizeOfRow || j >= sizeOfColumn)
		return False;
	*position = j * sizeOfRow + i;
	return True;
}

Boolean
NarrowSelection3D(Rubik3DWidget w, int *face, int *position, int *direction)
{
	switch (*direction) {
		case TOP:
		case RIGHT:
		case BOTTOM:
		case LEFT:
			if (w->rubik.vertical) {
				if (*face == 1 || *face == 3)
					return False;
			} else {
				if (*face == 0 || *face == 4)
					return False;
				if (*face == 5)
					*direction = (*direction + 2) % MAXORIENT;
			}
			break;
		case CCW:
		case CW:
			break;
		case TR:
			if (w->rubik.vertical) {
				if (*face == 0 || *face == 5)
					return False;
				else if (*face == 1 || *face == 2 || *face == 4)
					*direction = TOP;
				else	/* (*face == 3) */
					*direction = LEFT;
			} else {
				if (*face == 1 || *face == 5)
					return False;
				else if (*face == 0 || *face == 2 || *face == 3)
					*direction = RIGHT;
				else	/* (*face == 4) */
					*direction = BOTTOM;
			}
			break;
		case BR:
			if (w->rubik.vertical) {
				if (*face == 2 || *face == 4)
					return False;
				else if (*face == 0 || *face == 5)
					*direction = BOTTOM;
				else if (*face == 1)
					*direction = RIGHT;
				else	/* (*face == 3) */
					*direction = TOP;
			} else {
				if (*face == 2 || *face == 3)
					return False;
				else if (*face == 4 || *face == 5)
					*direction = LEFT;
				else if (*face == 0)
					*direction = BOTTOM;
				else	/* (*face == 1) */
					*direction = RIGHT;
			}
			break;
		case BL:
			if (w->rubik.vertical) {
				if (*face == 0 || *face == 5)
					return False;
				else if (*face == 1 || *face == 2 || *face == 4)
					*direction = BOTTOM;
				else	/* (*face == 3) */
					*direction = RIGHT;
			} else {
				if (*face == 1 || *face == 5)
					return False;
				else if (*face == 0 || *face == 2 || *face == 3)
					*direction = LEFT;
				else	/* (*face == 4) */
					*direction = TOP;
			}
			break;
		case TL:
			if (w->rubik.vertical) {
				if (*face == 2 || *face == 4)
					return False;
				else if (*face == 0 || *face == 5)
					*direction = TOP;
				else if (*face == 1)
					*direction = LEFT;
				else	/* (*face == 3) */
					*direction = BOTTOM;
			} else {
				if (*face == 2 || *face == 3)
					return False;
				else if (*face == 4 || *face == 5)
					*direction = RIGHT;
				else if (*face == 0)
					*direction = TOP;
				else	/* (*face == 1) */
					*direction = LEFT;
			}
			break;
		default:
			return False;
	}
	/* Remap to row movement */
	if (*direction == CW || *direction == CCW) {
		int         i, j, sizeOfRow, sizeOfColumn, newFace;

		newFace = rotateToRow[*face].face;
		faceSizes((RubikWidget) w, newFace, &sizeOfRow, &sizeOfColumn);
		*direction = (*direction == CCW) ?
			(rotateToRow[*face].direction + 2) % MAXORIENT :
			rotateToRow[*face].direction;
		if (rotateToRow[*face].sideFace == LEFT ||
		    rotateToRow[*face].sideFace == BOTTOM) {
			i = sizeOfRow - 1;
			j = sizeOfColumn - 1;
		} else {
			i = j = 0;
		}
		*face = newFace;
		*position = j * sizeOfRow + i;
	}
	return True;
}

static void
DrawFrame(Rubik3DWidget w, GC gc)
{
	int         face, dx, dy;

	dx = w->rubik3d.viewSize.x + w->rubik.puzzleOffset.x;
	dy = w->rubik3d.viewSize.y + w->rubik.puzzleOffset.y;
	if (w->rubik.vertical) {
		XDrawLine(XtDisplay(w), XtWindow(w), gc,
			  0, dy, dx + w->rubik.puzzleOffset.x + 1, dy);
		XDrawString(XtDisplay(w), XtWindow(w), gc,
			    (int) (2 * w->rubik.delta),
			(int) (3 * w->rubik.delta + w->rubik.letterOffset.y),
			    "Front", 5);
		XDrawString(XtDisplay(w), XtWindow(w), gc, (int)
			    (-4 * w->rubik.delta + 2 * 4 * w->rubik.letterOffset.x + w->core.width),
			    (int) (-w->rubik.delta - 2 * w->rubik.letterOffset.y + w->core.height),
			    "Back", 4);
	} else {
		XDrawLine(XtDisplay(w), XtWindow(w), gc,
			  dx, 0, dx, dy + w->rubik.puzzleOffset.y + 1);
		XDrawString(XtDisplay(w), XtWindow(w), gc,
			    (int) (2 * w->rubik.delta),
			(int) (3 * w->rubik.delta + w->rubik.letterOffset.y),
			    "Front", 5);
		XDrawString(XtDisplay(w), XtWindow(w), gc, (int)
			    (-4 * w->rubik.delta + 2 * 4 * w->rubik.letterOffset.x + w->core.width),
			    (int) (-w->rubik.delta - 2 * w->rubik.letterOffset.y + w->core.height),
			    "Back", 4);
	}
	for (face = 0; face < MAXFACES; face++)
		XDrawLines(XtDisplay(w), XtWindow(w), gc,
			   faceLoc3D[face], MAXORIENT, CoordModePrevious);
}

void
DrawSquare3D(Rubik3DWidget w, int face, int position, int offset)
{
	GC          faceGC, borderGC;
	int         x, y, dx, dy, i, j, sizeOfRow;

	sizeOfRow = sizeRow((RubikWidget) w, face);
	i = position % sizeOfRow;
	j = position / sizeOfRow;
	MapTo3D(w, face, i, j, &x, &y);
	CubeOffset3D(w, face, x, y, &dx, &dy);
	cubeLoc3D[face][0].x = dx;
	cubeLoc3D[face][0].y = dy;
	if (offset) {
		borderGC = w->rubik.faceGC[(int) w->rubik.cubeLoc[face][position].face];
		if (w->rubik.depth < 2 || w->rubik.mono) {
			faceGC = w->rubik.inverseGC;
		} else {
			faceGC = w->rubik.borderGC;
		}
	} else {
		faceGC = w->rubik.faceGC[(int) w->rubik.cubeLoc[face][position].face];
		borderGC = w->rubik.borderGC;
	}
	XFillPolygon(XtDisplay(w), XtWindow(w),
	      faceGC, cubeLoc3D[face], MAXORIENT, Convex, CoordModePrevious);
	XDrawLines(XtDisplay(w), XtWindow(w),
		   borderGC, cubeLoc3D[face], 5, CoordModePrevious);
	if (w->rubik.depth < 2 || w->rubik.mono) {
		int         letterX, letterY;
		char        buf[2];

		(void) sprintf(buf, "%c",
			       w->rubik.faceName[(int) w->rubik.cubeLoc[face][position].face][0]);
		letterX = dx + letter3DList[face].x + w->rubik.letterOffset.x;
		letterY = dy + letter3DList[face].y + w->rubik.letterOffset.y;
		if (offset) {
			borderGC = w->rubik.borderGC;
		} else {
			borderGC = w->rubik.inverseGC;
		}
		XDrawString(XtDisplay(w), XtWindow(w), borderGC,
			    letterX, letterY, buf, 1);
	}
	if (w->rubik.orient) {
		XDrawLine(XtDisplay(w), XtWindow(w), borderGC,
			  dx + orient3DList[face][(int) w->rubik.cubeLoc[face][position].rotation][0].x,
			  dy + orient3DList[face][(int) w->rubik.cubeLoc[face][position].rotation][0].y,
			  dx + orient3DList[face][(int) w->rubik.cubeLoc[face][position].rotation][1].x,
			  dy + orient3DList[face][(int) w->rubik.cubeLoc[face][position].rotation][1].y);
	}
}

static void
MapTo3D(Rubik3DWidget w, int face, int i, int j, int *x, int *y)
{
	int         sizeOfRow, sizeOfColumn;

	faceSizes((RubikWidget) w, face, &sizeOfRow, &sizeOfColumn);
	switch (face) {
		case 0:
			*x = w->rubik.sizez - 1 - j;
			*y = i;
			break;
		case 1:
			*x = j;
			*y = w->rubik.sizez - 1 - i;
			break;
		case 2:
			*x = i;
			*y = j;
			break;
		case 3:
			*x = w->rubik.sizez - 1 - i;
			*y = w->rubik.sizey - 1 - j;
			break;
		case 4:
			*x = w->rubik.sizex - 1 - i;
			*y = w->rubik.sizez - 1 - j;
			break;
		case 5:
			*x = j;
			*y = w->rubik.sizex - 1 - i;
			break;
		default:
			(void) printf("MapTo3D: face %d\n", face);
	}
}

#ifdef DEBUG
static void
MapFrom3D(w, face, x, y, i, j)
	Rubik3DWidget w;
	int         face, x, y, *i, *j;
{
	switch (face) {
		case 0:
			*i = y;
			*j = w->rubik.size - 1 - x;
			break;
		case 1:
			*i = w->rubik.size - 1 - y;
			*j = x;
			break;
		case 2:
			*i = x;
			*j = y;
			break;
		case 3:
			*i = w->rubik.size - 1 - x;
			*j = w->rubik.size - 1 - y;
			break;
		case 4:
			*i = w->rubik.size - 1 - x;
			*j = w->rubik.size - 1 - y;
			break;
		case 5:
			*i = w->rubik.size - 1 - y;
			*j = x;
			break;
		default:
			(void) printf("MapFrom3D: face %d\n", face);
	}
}

#endif

static void
MapOrientFrom3D(int face, int corner, int *side)
{
	switch (face) {
		case 0:
			*side = (corner + 2) % MAXORIENT;
			break;
		case 1:
			*side = corner;
			break;
		case 2:
			*side = (corner + 3) % MAXORIENT;
			break;
		case 3:
			*side = (corner + 1) % MAXORIENT;
			break;
		case 4:
			*side = (corner + 1) % MAXORIENT;
			break;
		case 5:
			*side = corner;
			break;
		default:
			(void) printf("MapFrom3D: face %d\n", face);
	}
}

static void
CubeOffset3D(Rubik3DWidget w, int face, int x, int y, int *dx, int *dy)
{
	if (w->rubik.vertical)
		switch (face) {
			case 0:
				*dx = w->rubik3d.viewMiddle.x + w->rubik.delta - 1 +
					y * (w->rubik3d.cubeSizex.x + w->rubik.delta) -
					x * (w->rubik3d.cubeDiagonalz + w->rubik.delta);
				*dy = w->rubik3d.viewMiddle.y - w->rubik.delta - 2 -
					x * (w->rubik3d.cubeSizez.y + w->rubik.delta);
				break;
			case 1:
				*dx = w->rubik3d.viewMiddle.x - 2 * w->rubik.delta -
					x * (w->rubik3d.cubeDiagonaly + w->rubik.delta) -
					y * (w->rubik3d.cubeDiagonalz + w->rubik.delta);
				*dy = w->rubik3d.viewMiddle.y +
					x * (w->rubik3d.cubeSizey.y + w->rubik.delta) -
					y * (w->rubik3d.cubeSizez.y + w->rubik.delta);
				break;
			case 2:
				*dx = w->rubik3d.viewMiddle.x + w->rubik.delta - 1 +
					x * (w->rubik3d.cubeSizex.x + w->rubik.delta) -
					y * (w->rubik3d.cubeDiagonaly + w->rubik.delta);
				*dy = w->rubik3d.viewMiddle.y + 2 * w->rubik.delta - 1 +
					y * (w->rubik3d.cubeSizey.y + w->rubik.delta);
				break;
			case 3:
				*dx = w->rubik3d.viewMiddle.x + 2 * w->rubik.delta +
					x * (w->rubik3d.cubeDiagonalz + w->rubik.delta) +
					y * (w->rubik3d.cubeDiagonaly + w->rubik.delta);
				*dy = w->rubik3d.viewSize.y + w->rubik3d.viewMiddle.y +
					w->rubik.delta - 1 -
					x * (w->rubik3d.cubeSizez.y + w->rubik.delta) +
					y * (w->rubik3d.cubeSizey.y + w->rubik.delta);
				break;
			case 4:
				*dx = w->rubik3d.viewMiddle.x - w->rubik.delta + 1 -
					x * (w->rubik3d.cubeSizex.x + w->rubik.delta) +
					y * (w->rubik3d.cubeDiagonalz + w->rubik.delta);
				*dy = w->rubik3d.viewSize.y + w->rubik3d.viewMiddle.y -
					w->rubik.delta + 1 -
					y * (w->rubik3d.cubeSizez.y + w->rubik.delta);
				break;
			case 5:
				*dx = w->rubik3d.viewMiddle.x - 2 -
					y * (w->rubik3d.cubeSizex.x + w->rubik.delta) +
					x * (w->rubik3d.cubeDiagonaly + w->rubik.delta);
				*dy = w->rubik3d.viewSize.y + w->rubik3d.viewMiddle.y +
					2 * w->rubik.delta + 2 +
					x * (w->rubik3d.cubeSizey.y + w->rubik.delta);
				break;
			default:
				(void) printf("CubeOffset3D: face %d\n", face);
	} else
		switch (face) {
			case 0:
				*dx = w->rubik3d.viewMiddle.x +
					y * (w->rubik3d.cubeSizex.x + w->rubik.delta) -
					x * (w->rubik3d.cubeSizez.x + w->rubik.delta);
				*dy = w->rubik3d.viewMiddle.y - 2 * w->rubik.delta + 1 -
					y * (w->rubik3d.cubeDiagonalx + w->rubik.delta) -
					x * (w->rubik3d.cubeDiagonalz + w->rubik.delta);
				break;
			case 1:
				*dx = w->rubik3d.viewMiddle.x - w->rubik.delta - 2 -
					y * (w->rubik3d.cubeSizez.x + w->rubik.delta);
				*dy = w->rubik3d.viewMiddle.y + w->rubik.delta +
					x * (w->rubik3d.cubeSizey.y + w->rubik.delta) -
					y * (w->rubik3d.cubeDiagonalz + w->rubik.delta);
				break;
			case 2:
				*dx = w->rubik3d.viewMiddle.x + 2 * w->rubik.delta - 1 +
					x * (w->rubik3d.cubeSizex.x + w->rubik.delta);
				*dy = w->rubik3d.viewMiddle.y + w->rubik.delta +
					y * (w->rubik3d.cubeSizey.y + w->rubik.delta) -
					x * (w->rubik3d.cubeDiagonalx + w->rubik.delta);
				break;
			case 3:
				*dx = w->rubik3d.viewSize.x + w->rubik3d.viewMiddle.x -
					w->rubik.delta + 1 -
					x * (w->rubik3d.cubeSizez.x + w->rubik.delta);
				*dy = w->rubik3d.viewMiddle.y - w->rubik.delta -
					y * (w->rubik3d.cubeSizey.y + w->rubik.delta) +
					x * (w->rubik3d.cubeDiagonalz + w->rubik.delta);
				break;
			case 4:
				*dx = w->rubik3d.viewSize.x + w->rubik3d.viewMiddle.x +
					w->rubik.delta - 1 -
					y * (w->rubik3d.cubeSizez.x + w->rubik.delta) +
					x * (w->rubik3d.cubeSizex.x + w->rubik.delta);
				*dy = w->rubik3d.viewMiddle.y + 2 * w->rubik.delta +
					y * (w->rubik3d.cubeDiagonalz + w->rubik.delta) +
					x * (w->rubik3d.cubeDiagonalx + w->rubik.delta);
				break;
			case 5:
				*dx = w->rubik3d.viewSize.x + w->rubik3d.viewMiddle.x +
					2 * w->rubik.delta + 2 +
					y * (w->rubik3d.cubeSizex.x + w->rubik.delta);
				*dy = w->rubik3d.viewMiddle.y - w->rubik.delta -
					x * (w->rubik3d.cubeSizey.y + w->rubik.delta) +
					y * (w->rubik3d.cubeDiagonalx + w->rubik.delta);
				break;
			default:
				(void) printf("CubeOffset3D: face %d\n", face);
		}
}
